/* (C) Dan S. Reznik - UCB EECS */

#define VERSION_INFO "Firmware V-1.0 - (C) Dan Reznik, 2/1/2000"

#define ISR_FREQ 5000

#define CYCLE_SAMPLES 64
#define CYCLE_MASK (CYCLE_SAMPLES-1)

/* TODO: Make OP_Set_* change value during chirping or off phase */

typedef enum {
  ON_MODE_SUSTAIN,ON_MODE_PULSE_ONCE,ON_MODE_PULSE_TRAIN
} enum_ON_MODE;

typedef enum {
  ENV_INIT,
  ENV_ATTACK_1,ENV_ATTACK_2,ENV_SUSTAIN,ENV_DECAY_1,ENV_DECAY_2,
  ENV_PAUSE,
  ENV_OFF
} enum_ENV;

typedef enum {
  SHIFT_UP_MID,
  SHIFT_DN_MID,
  SHIFT_UP_TOP,
  SHIFT_DN_TOP,
  SHIFT_NO_ENV
} enum_SHIFT;

typedef enum {
  OP_On,
  OP_Off,
  OP_Inc,          /* echoes incremented byte */
  OP_Sample,       /* gets 4 11-bit samples */
  OP_Sample_Batch, /* gets many sample groups */
  OP_Sample_Batch_Sync, /* 1st sample @ k_index overflow */
  OP_Set_Envelope, /* attack, sustain, pause, shift, shift_mode */
  OP_Set_Wave_T,   /* Translation */
  OP_Set_Wave_R,   /* Rotation */
  OP_Set_Steps,    /* step_1, step_2 */
  OP_Set_Debug,    /* debug_1, debug_2 */
  OP_Set_Freq,
  OP_Set_On_Mode,  /* on_mode */
  OP_Get_State,    /* envelope state */    
  OP_Get_Glob,
  OP_Get_Wave_T,
  OP_Get_Wave_R,
  OP_Get_Version_Info,
  OP_NUM
} enum_OP;

typedef struct {
  unsigned short stamp;
  signed short step_1, step_2;
  unsigned short attack, sustain, pause; /* in 1/ISR_FREQ units */
  /* list all even-sized globals above to avoid alignment problems */
  unsigned char shift, /* enum_SHIFT */ shift_mode;
  unsigned char state; /* enum_ENV */
  unsigned char on_mode; /* enum_ON_MODE */
  unsigned char freq; /* actual_freq = (ISR_FREQ/CYCLE_SAMPLES)*(freq+1)/256 */
} struct_glob;
