#include <pic.h>
#include "firmware.h"

typedef struct {
  unsigned char char_0, char_1;
} _short_char;

typedef struct {
  unsigned char char_0;
  unsigned short short_mid;
  unsigned char char_3;
} _long_mid;

typedef struct {
  unsigned char char_0;
  unsigned char char_1;
  unsigned short short_top;
} _long_top;

/* 10-bit DC = CCPR1L[8] : CCP1X[1] : CCP1Y[1] */

#define SET_DUTY(/* unsigned short */ dc1, /* unsigned short */ dc2) \
  CCP1Y = (dc1&0x01?1:0); \
  CCP1X = (dc1&0x02?1:0); \
  CCPR1L = (char)((dc1>>2)&0xff); \
  CCP2Y = (dc2&0x01?1:0); \
  CCP2X = (dc2&0x02?1:0); \
  CCPR2L = (char)((dc2>>2)&0xff)

#define DEBUG_PORTC(/* char */ bcd) \
  PORTC |= (bcd&0x8?0x80:0) | (bcd&0x4?0x40:0) | \
           (bcd&0x2?0x20:0) | (bcd&0x1?0x08:0);

#define Set_CONVST(/* bit */ val) (val ?  PORTA |= 0x10 : PORTA &= 0x2f)
#define Get_BUSY() (PORTA&0x01)
#define Wait_For_BUSY() { while (!Get_BUSY()) ; }

#define MAKE_15_BIT_SIGNED(/* signed short */ dc) \
{ if (dc < 0) dc = -(((-dc)>>1)&0x7fff); else dc = (dc>>1)&0x7fff; }

#define GET_ENV8(/* ulong */ env_val, /* char */ env_8) \
  switch(glob.shift_mode) { \
    case SHIFT_NO_ENV: \
      env_8 = 255; \
      break; \
    case SHIFT_DN_TOP: \
      short_aux = ((_long_top *)&env_val)->short_top; \
      short_aux >>= glob.shift; \
      env_8 = ((_short_char *)&short_aux)->char_0; \
      break; \
      case SHIFT_UP_TOP: \
      short_aux = ((_long_top *)&env_val)->short_top; \
      short_aux <<= glob.shift; \
      env_8 = ((_short_char *)&short_aux)->char_1; \
      break; \
    case SHIFT_DN_MID: \
      short_aux = ((_long_mid *)&env_val)->short_mid; \
      short_aux >>= glob.shift; \
      env_8 = ((_short_char *)&short_aux)->char_0; \
      break; \
    case SHIFT_UP_MID: \
      short_aux = ((_long_mid *)&env_val)->short_mid; \
      short_aux <<= glob.shift; \
      env_8 = ((_short_char *)&short_aux)->char_1; \
      break; \
  }

#define SIGN_SPLIT(/* short */ signed_val, /* char */ sign, \
                   /* unsigned short */ unsigned_val) \
  if (signed_val < 0) { \
    sign = 1; unsigned_val = -signed_val; \
  } else { \
    sign = 0; unsigned_val = signed_val; \
  }

extern struct_glob glob;
extern unsigned char k_index;
extern bank2 signed char glob_table_T[CYCLE_SAMPLES];
extern bank3 signed char glob_table_R[CYCLE_SAMPLES];

const static enum_ENV glob_new_state[] = {
  ENV_ATTACK_1,  // ENV_INIT,
  ENV_ATTACK_2,  // ENV_ATTACK_1,
  ENV_SUSTAIN,   // ENV_ATTACK_2,
  ENV_DECAY_1,   // ENV_SUSTAIN,
  ENV_DECAY_2,   // ENV_DECAY_1,
  ENV_PAUSE,     // ENV_DECAY_2,
  ENV_OFF,       // ENV_PAUSE,
  ENV_OFF        // ENV_OFF
};

#define ONE_15 0x8000 /* 1<<15 */

void interrupt isr(void) {
  static unsigned short step_1, step_2, linear_1, linear_2;
  static signed short dc_scaled_1, dc_scaled_2, dc_combo;
  static unsigned short iterations_to_go;
  static unsigned short T_plus_R, T_minus_R;
  static unsigned long squared_1, squared_2;
  static signed char dc_sample;
  static unsigned char env_8, sign_1, sign_2;
  static unsigned short k;
  static unsigned short short_aux;

  if (TMR2IF) {
    TMR2IF = 0;
    Set_CONVST(1);

    if (glob.state == ENV_OFF) {
      SET_DUTY(1<<9,1<<9); /* half way, avoids transient */
    } else {
      if (glob.state == ENV_INIT) {
	iterations_to_go = 0; /* so jumps into switch() below */
	linear_1 = 0; squared_1 = 0;
	SIGN_SPLIT(glob.step_1, sign_1, step_1);
	linear_2 = 0; squared_2 = 0;
	SIGN_SPLIT(glob.step_2, sign_2, step_2);
	k = 0;
      }
    
      k_index = ((_short_char *)&k)->char_1;
      k_index &= CYCLE_MASK;

      dc_sample = glob_table_T[k_index];
      GET_ENV8(squared_1, env_8);
      dc_scaled_1 = (sign_1 ? -dc_sample : dc_sample) * env_8;
      MAKE_15_BIT_SIGNED(dc_scaled_1);

      dc_sample = glob_table_R[k_index];
      GET_ENV8(squared_2, env_8);
      dc_scaled_2 = (sign_2 ? -dc_sample : dc_sample) * env_8;
      MAKE_15_BIT_SIGNED(dc_scaled_2);

      dc_combo  = dc_scaled_1 + dc_scaled_2; /* 16 bit signed */
      T_plus_R  = ((dc_combo + ONE_15)>>6)&0x3ff; /* 10 bit */
      dc_combo  = dc_scaled_1 - dc_scaled_2; /* 16 bit signed */
      T_minus_R = ((dc_combo + ONE_15)>>6)&0x3ff; /* 10 bit */
      SET_DUTY(T_plus_R, T_minus_R);

      k += glob.freq; k++; /* increment by freq+1 */
    
      if ((glob.on_mode != ON_MODE_SUSTAIN) || (glob.state != ENV_SUSTAIN))
	if (!iterations_to_go) {
	  if ((glob.on_mode==ON_MODE_PULSE_TRAIN)&&(glob.state==ENV_PAUSE))
	    glob.state = ENV_ATTACK_1; /* CHIRP IT */
	  else
	    glob.state = glob_new_state[glob.state];
	  switch(glob.state) {
	     case ENV_SUSTAIN : iterations_to_go = glob.sustain; break;
	     case ENV_PAUSE   : iterations_to_go = glob.pause; break;
	     default : iterations_to_go =  glob.attack>>1; break;
	  }
	}
  
      switch(glob.state) {
      case ENV_ATTACK_1:
	linear_1 += step_1; squared_1 += linear_1;
	linear_2 += step_2; squared_2 += linear_2;
	break;
      case ENV_ATTACK_2:
	linear_1 -= step_1; squared_1 += linear_1;
	linear_2 -= step_2; squared_2 += linear_2;
	break;
      case ENV_DECAY_1:
	linear_1 += step_1; squared_1 -= linear_1;
	linear_2 += step_2; squared_2 -= linear_2;
	break;
      case ENV_DECAY_2:
	linear_1 -= step_1; squared_1 -= linear_1;
	linear_2 -= step_2; squared_2 -= linear_2;
	break;
      }
      iterations_to_go--;
    }

    Set_CONVST(0); /* Wait_For_BUSY(); */
    glob.stamp++; /* time stamp it */
  }
}
